//* AnalysIR Firmware for Photon via Serial & WiFi
//* Date: 25th October 2015
//* Release: 1.0.x
//* Note: Please use and read in conjunction with the README and the AnalysIR Getting Started Guide.
//* Licence: Free to use & modify without restriction or warranty for non-commercial uses.
//*          Free to use & modify without restriction or warranty for commercial uses with a registered copy of AnalysIR.
//*          For other uses please contact the Author directly.
//*          Please acknowledge AnalysIR as the Author in any derivative and include a link to http://www.AnalysIR.com in any publication.
//*
//* Tested with Particle Web IDE Photon Firmware 0.4.7

//set true for Micro-controller system being used, set all others to false
#define Photon true    //set to true for Photon   
#define THREADING true  // set true to enalbe Photon Threading or false to disable it
#if THREADING
SYSTEM_THREAD(ENABLED); //run user (this) code in its own 'thread'
#endif
STARTUP( startup_Photon() ); //must be done ASAP to avoid turning on IR emitter by accident

SYSTEM_MODE(MANUAL);  //we don't need automatic mode for this application



//Definitions for Buffers here...size RAM dependent
#define modPULSES 256 //increase until memory is used up, max 256, leave at 256.

//General Definitions

//Baud rate is now fixed to 115200 for all devices, to avoid issues with some platforms
#define BAUDRATE 115200

#if Photon
#define AVR8BITMCU false
#define IR_Rx_PIN D2
#define IR_Mod_PIN D3
#define ledPin    D7
#define pin2HIGH pinReadFast(IR_Rx_PIN)
#define maxPULSES 1024 //More RAM is available on this Photon
#else
ERROR - you must define Photon as true above
#endif

#define NETDEBUG false   //set true to print net traffic on serial as well. Otherwise nothing on serial when net link is active
#if NETDEBUG
#define    netDebugPrint(x) Serial.print(x)
#define    netDebugPrintln(x) Serial.println(x)
#else
#define    netDebugPrint(x)   // do nothing...
#define    netDebugPrintln(x)   // do nothing...
#endif//...............................................................................................

unsigned short pulseIR[maxPULSES]; //temp store for pulse durations (demodulated)
volatile byte modIR[modPULSES]; //temp store for modulation pulse durations - changed in interrupt

//General variables
volatile byte state = 127; //defines initial value for state normally HIGH or LOW. Set in ISR.
byte oldState = 127; //set both the same to start. Used to test for change in state signalled from ISR
unsigned long usLoop, oldTime; //timer values in uSecs. usLoops stores the time value of each loop, oldTime remembers the last time a state change was received
volatile unsigned long newMicros;//passes the time a state change occurred from ISR to main loop
unsigned long oldMicros = 0; //passes the time a state change occurred from ISR to main loop
unsigned short countD = 0; // used as a pointer through the buffers for writing and reading (de-modulated signal)
volatile byte countM = 0; // used as a pointer through the buffers for writing and reading (modulated signal)
byte countM2 = 0; //used as a counter, for the number of modulation samples used in calculating the period.
unsigned short i = 0; //used to iterate in for loop...integer
byte j = 0; //used to iterate in for loop..byte
unsigned int sum = 0; //used in calculating Modulation frequency
byte sigLen = 0; //used when calculating the modulation period. Only a byte is required.

//Serial Tx buffer - uses Serial.write for faster execution
byte txBuffer[5]; //Key(+-)/1,count/1,offset/4,CR/1   <= format of packet sent to AnalysIR over serial

// we use smtp Port - defaults to port 25 (Using this port means less problems with firewalls etc)
TCPServer server = TCPServer(25);
TCPClient client;
char netStringBuf [maxPULSES * 8]; //used to build up string for each line
const short  netStringBufLen = sizeof(netStringBuf);
short netStringPtr = 0; //pointer for next free position in netStringBuf


void setup() {

  Serial.begin(BAUDRATE);//fixed at 115200 bps for all platforms

  Particle.connect();             // non-blocking attempt to connect to Wifi
  for (i = 0; i < 20; i++) { //wait for new flash instructions from cloud
    digitalWriteFast(ledPin,!digitalRead(ledPin)); // animate start-up process
    if (!Particle.connected) Particle.connect();
    Particle.process();
    delay(1000);   // wait 20 seconds at startup
  }
  if (Particle.connected) Particle.disconnect();             // no more cloud

  server.begin(); //listen for client connections via WiFi from AnalysIR

  txBuffer[4] = 13; //init ascii decimal value for CR in tx buffer (used for Serial only)

  digitalWrite(ledPin, HIGH); //blink Led briefly
  delay(500);//time to see blink
  digitalWrite(ledPin, LOW);

  Serial.println(F("!AnalysIR http://www.AnalysIR.com/ !")); // HELLO STRING - ALL COMMENTS SENT IN !....! FORMAT

  //Print Hello
  Serial.println(F("!Photon over Serial or WiFi!")); // HELLO STRING - ALL COMMENTS SENT IN !....! FORMAT

  //Print IPaddress
  Serial.print(F("!IP Address ")); Serial.print(WiFi.localIP()); Serial.println(F(" !")); //dump WiFI IP address to Serial

  //Print SSID
  Serial.print(F("!SSID ")); Serial.print(WiFi.SSID()); Serial.println(F(" !")); //dump SSID to Serial

  //Print RSSI
  Serial.print(F("!RSSI ")); Serial.print(WiFi.RSSI()); Serial.println(F(" !")); //dump RSSI to Serial

  //Print network buffer length
  Serial.print(F("!Network Buffer Length ")); Serial.print(netStringBufLen); Serial.println(F(" !")); //dump WiFI IP address to Serial

  //Initialise State
  oldState = digitalRead(IR_Rx_PIN);
  state = oldState;

  //Initialise Times
  oldTime = 0; //initialise
  newMicros = micros(); //initialise
  oldMicros = newMicros;

  //following line not required - just reports free RAM on Arduino if there are problems
  reportPhotonFreeRAM(0xFFFF);// report free ram to host always, use max UInt value of 0xFFFF. 8Bit AVRs & Photon only
  //

  //turn on interrupts and GO!
  attachInterrupt(IR_Rx_PIN, rxIR_Interrupt_Handler, CHANGE);
  attachInterrupt(IR_Mod_PIN, rxIR3_Interrupt_Handler, FALLING);
}

void loop() {
  while (true) { //avoid any extra stuff inserted by Arduino IDE at end of each normal Loop

    usLoop = micros(); //used once every loop rather than multiple calls

    if (oldState != state && countD < maxPULSES) {
      oldState = state;
      if (oldState) { //if the duration is longer than 0xFFFF(65535 uSecs) then multiple repeat pulses are stored, with LSB used to signal mark or space

        sum = (newMicros - oldMicros); //re-use sum var here, to save RAM (normally used in reportperiod)

        //while (sum > 0xFFFF && countD < (maxPULSES - 1) && countD) { //this allows for a mark/space of greater than 65535 uSecs (0xFFFF), ignore first signal
        if (sum > 0xFFFF && countD < (maxPULSES - 1) && countD) { //this allows for a mark/space of greater than 65535 uSecs (0xFFFF), ignore first signal
          //sum -= 0xFFFF;//this assumes the length is not longer than 131070
          pulseIR[countD++] = 0xFFFF | 0x0001; //store for later & include state
        }
        pulseIR[countD++] = (sum & 0xFFFF) | 0x0001; //store for later & include state
      }
      else {
        sum = (newMicros - oldMicros); //re-use sum var here, to save RAM (normally used in reportperiod)
        //while (sum > 0xFFFF && countD < (maxPULSES - 1) && countD) { //this allows for a mark/space of greater than 65535 uSecs (0xFFFF), ignore first signal
        if (sum > 0xFFFF && countD < (maxPULSES - 1) && countD) { //this allows for a mark/space of greater than 65535 uSecs (0xFFFF), ignore first signal
          //sum -= 0xFFFF;//this assumes the length is not longer than 131070
          pulseIR[countD++] = 0xFFFF & 0xFFFE; //store for later & include state
        }
        pulseIR[countD++] = sum & 0xFFFE; //store for later & include state
      }
      oldMicros = newMicros; //remember for next time
      oldTime = usLoop; //last time IR was received
    }


    if (state && countD > 0 && (countD == maxPULSES ||  (usLoop - oldTime) > 125000UL)) { //if we have received maximum pulses or its 100ms since last one
      detachInterrupt(IR_Rx_PIN);
      detachInterrupt(IR_Mod_PIN);

      reportPulses();
      reportPeriod();//reports modulation frequency to host over serial
      countD = 0; //reset value for next time

      attachInterrupt(IR_Rx_PIN, rxIR_Interrupt_Handler, CHANGE);
      attachInterrupt(IR_Mod_PIN, rxIR3_Interrupt_Handler, FALLING);
    }

  }
}

void  reportPulses() {
  if (!client.connected()) client = server.available();

  if (client.connected()) { //workaround to detect remotely closed net connection
    Serial.print(F("flush/read "));
    short x = client.read();
    Serial.println(x);
    //client.flush();
    while (client.available()) client.read(); //read & discard  anything available from network
  }


  if (client.connected()) { //if AnalysIR connected via net
    //blink LED if network connection
    digitalWrite(D7, HIGH); //force High for net - brief blink
    client.read(); //read and ignore any incomming traffic
    netStringPtr = 0;
    for (i = 0; i < countD; i++) {
      //the following logic takes care of the inverted signal in the IR receiver
      if (pulseIR[i] & 0x01) { //mark or pulse
        netStringPtr += sprintf(netStringBuf + netStringPtr, "+%d\r\n", pulseIR[i]);
      }
      else {  //space
        netStringPtr += sprintf(netStringBuf + netStringPtr, "-%d\r\n", pulseIR[i]);
      }
      
      if (netStringPtr > (netStringBufLen - 20)) {
        unsigned int x = client.print(netStringBuf);
        Serial.print(F("netStringBuf Full #1 sent length:")); Serial.println(x);
        netStringPtr = 0;
      }
    }
    
    if (netStringPtr > (netStringBufLen - 40)) { //only send if not enough room left for carrier freq later
      unsigned short x = client.print(netStringBuf);
      Serial.print(F("netStringBuf Full #2 sent length: ")); Serial.println(x);
      netStringPtr = 0;
    }
    digitalWrite(D7, LOW); //brief pulse for net connection

  } else { //AnalysIR connected via serial
    //toggle LED if no network connection
    digitalWrite(D7, !digitalRead(D7)); //toggle every signal on serial

    for (i = 0; i < countD; i++) {
      //the following logic takes care of the inverted signal in the IR receiver
      if (pulseIR[i] & 0x01) txBuffer[0] = '+'; //Mark is sent as +...LSB bit of pulseIR is State(Mark or Space)
      else txBuffer[0] = '-';           //Space is sent as -
      txBuffer[1] = (byte) (i & 0xFF); //count
      txBuffer[3] = pulseIR[i] >> 8; //byte 1
      txBuffer[2] = pulseIR[i] & 0xFE; //LSB 0 ..remove last bit as it was State
      Serial.write(txBuffer, 5);
    }
  }
}

void  reportPeriod() { //report period of modulation frequency in nano seconds for more accuracy.
  sum = 0; // UL
  sigLen = 0; //byte
  countM2 = 0; //byte
  for (j = 1; j < (modPULSES - 1); j++) { //i is byte
    sigLen = (modIR[j] - modIR[j - 1]); //siglen is byte
    if (sigLen > 50 || sigLen < 10) continue; //this is the period range length exclude extraneous ones
    sum += sigLen; // sum is UL
    countM2++; //countM2 is byte
    modIR[j - 1] = 0; //finished with it so clear for next time
  }
  modIR[j - 1] = 0; //now clear last one, which was missed in loop
  if (countM2 == 0) return; //avoid div by zero = nothing to report
  sum =  (sum * 1000 + countM2/2) / countM2; //get it in nano secs
  // now send over serial using buffer
  txBuffer[0] = 'M'; //Modulation report is sent as 'M'
  txBuffer[1] = countM2; //number of samples used
  txBuffer[3] = sum >> 8 & 0xFF; //byte Period MSB
  txBuffer[2] = sum & 0xFF; //byte Period LSB
  if (client.connected()) { //if AnalysIR connected via net
    //client.read(); //read and ignore any incoming traffic
    netStringPtr += sprintf(netStringBuf + netStringPtr,  "! Carrier Frequency %d !\r\n", sum);
    netStringPtr += sprintf(netStringBuf + netStringPtr,  "! Signal End !\r\n");
    unsigned short x = client.write(netStringBuf);
    Serial.print(F("netStringBuf Sent in report period: ")); Serial.println(x);
    Serial.print(F("netStringBuf Sent in report period #2: ")); Serial.println(netStringPtr);
    netStringPtr = 0;
  } 
  else {
    Serial.write(txBuffer, 5);
    // if no client is yet connected, check for a new connection
    client = server.available();
  }
  return;
}

void rxIR_Interrupt_Handler() { //important to use few instruction cycles here
  //digital pin 2 on Arduino
  newMicros = micros(); //record time stamp for main loop
  state = pin2HIGH; //read changed state of interrupt pin 2 (return 4 or 0 for High/Low)

}

void rxIR3_Interrupt_Handler() { //important to use few instruction cycles here
  //digital pin 3 on Arduino - FALLING edge only
  modIR[countM++] = micros(); //just continually record the time-stamp, will be mostly modulations
  //just save LSB as we are measuring values of 20-50 uSecs only - so only need a byte (LSB)
}

void reportPhotonFreeRAM(unsigned int f) { //report freeRam to host if less than f or f is false
  uint32_t freemem = System.freeMemory();
  Serial.print(F("!Free RAM: "));//send line as comment
  Serial.print(freemem); // at least this amount of RAM available
  Serial.println(F("!"));//send line as comment
}

void startup_Photon() {
  //this should be called at startup to ensure any IR emitter or other connected devices are not turned on inadvertently, before Photon if fully initialised
  //for example.....
  pinMode(IR_Rx_PIN, INPUT);
  pinMode(IR_Mod_PIN, INPUT);
  pinMode(ledPin, OUTPUT);
  pinMode(TX, OUTPUT);
  digitalWrite(TX, HIGH);//make sure IR emitter(on A.IR Shield  is off initially
  //    digitalWrite(TX, LOW);//make sure IR emitter is off initially

}